#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_leylab_pipelines
----------------------------------

Tests for `leylab_pipelines` module.
"""


import sys
import unittest

from leylab_pipelines import leylab_pipelines



class TestLeylab_pipelines(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_someting(self):
        pass
